import matplotlib.pyplot as plt

from outset import draw_marquee


def test_draw_marquee_one():
    draw_marquee(
        frame_xlim=(0, 1),
        frame_ylim=(0, 2),
        color="mediumpurple",
        clip_on=False,
        despine=True,
        frame_edge_kwargs=dict(linewidth=0.5),
        leader_stretch=0.14,
        mark_glyph_kwargs=dict(markersize=15),
        leader_edge_kwargs=dict(linestyle=":", linewidth=2),
    )
    outpath = "/tmp/test_draw_marquee_one.png"
    plt.savefig(outpath)
    print(f"saved graphic to {outpath}")


def test_draw_marquee_no_leader():
    draw_marquee(
        frame_xlim=(0, 1),
        frame_ylim=(0, 2),
        color="mediumpurple",
        leader_stretch=0.0,
    )
    outpath = "/tmp/test_draw_marquee_no_leader.png"
    plt.savefig(outpath)
    print(f"saved graphic to {outpath}")


def test_draw_marquee_several():
    _fig, ax = plt.subplots(figsize=(6, 4))

    # Set limits and hide axes for a clean presentation
    ax.set_xlim(0, 4)
    ax.set_ylim(0, 4)

    draw_marquee(
        frame_xlim=[1, 1.25],
        frame_ylim=(0.5, 1.5),
        ax=ax,
    )
    draw_marquee(
        frame_xlim=[2, 3.9],
        frame_ylim=(0.5, 1.5),
        ax=ax,
    )
    draw_marquee(
        frame_xlim=[1, 2],
        frame_ylim=(2, 3),
        ax=ax,
    )

    # Display the updated plot
    outpath = "/tmp/test_draw_marquee_several.png"
    plt.savefig(outpath)
    print(f"saved graphic to {outpath}")



import pytest

from outset._auxlib.get_vertices_extent_ import get_vertices_extent


def test_get_vertices_extent():
    # Test with a simple rectangle
    vertices = [(0, 0), (2, 0), (2, 3), (0, 3)]
    expected_extent = (0, 2, 0, 3)
    assert get_vertices_extent(vertices) == expected_extent

    # Test with a single point
    vertices = [(1, 1)]
    expected_extent = (1, 1, 1, 1)
    assert get_vertices_extent(vertices) == expected_extent

    # Test with an empty list
    vertices = []
    with pytest.raises(ValueError):
        get_vertices_extent(vertices)



import typing

from frozendict import frozendict
import numpy as np

from matplotlib.axes import Axes as mpl_Axes
from matplotlib.collections import PathCollection as mpl_PathCollection
from matplotlib.container import ErrorbarContainer as mpl_ErrorbarContainer

from ._auxlib.align_marker_ import align_marker
from ._auxlib.rotate_marker_ import rotate_marker


def stub_edge_mark(
    ax: mpl_Axes,
    x: float,
    y: float,
    marker_kwargs: typing.Dict = frozendict(),
    offset: float = 0.1,
) -> typing.Tuple[float, float]:
    """Draw an edge marker for a single clipped point.

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        The axes object to draw markers on.
    x, y : float
        The x and y coordinates of the point.
    marker_kwargs : dict, optional
        Keyword arguments forwarded to matplotlib `plot` for edge markers.
    offset : float, default 0.1
        How far outside axis viewport to draw edge markers, proportional to
        axis viewport height or width.

    Returns
    -------
    edge_x, edge_y : float
        The x and y coordinates of the placed edge marker.

    See Also
    --------
    stub_all_clipped_values : Automates out of bounds scatterpoint detection
        and stub creation.
    """
    xlim, ylim = ax.get_xlim(), ax.get_ylim()
    edge_x, edge_y, markers = x, y, list()

    left_marker = lambda amount: align_marker(
        rf"$\langle\!\langle\!\langle | \!\! \leftrightarrow \!\!|{{\times}}{amount}$",
        halign="left",
        pad=1.3,
    )
    right_marker = lambda amount: align_marker(
        rf"$| \!\! \leftrightarrow \!\!|{{\times}}{amount}\rangle\!\rangle\!\rangle$",
        halign="right",
        pad=1.3,
    )

    xwidth = np.ptp(xlim)
    ywidth = np.ptp(ylim)
    xoff = xwidth * offset
    yoff = ywidth * offset
    if x < xlim[0]:
        amount = int((xlim[0] - x) / xwidth)
        edge_x = xlim[0] - xoff
        markers.append(left_marker(amount))
    elif x > xlim[1]:
        amount = int((x - xlim[1]) / xwidth)
        edge_x = xlim[1] + xoff
        markers.append(right_marker(amount))
    if y < ylim[0]:
        amount = int((ylim[0] - y) / ywidth)
        edge_y = ylim[0] - yoff
        markers.append(rotate_marker(right_marker(amount), 270))
    elif y > ylim[1]:
        amount = int((y - ylim[1]) / ywidth)
        edge_y = ylim[1] + yoff
        markers.append(rotate_marker(right_marker(amount), 90))

    for marker in markers:
        ax.plot(
            edge_x,
            edge_y,
            **{
                "color": "none",
                "clip_on": False,
                "linewidth": 0,
                "color": "red",
                "marker": marker,
                "markersize": 100,
                "zorder": 10000,
                **marker_kwargs,
            },
        )

    return edge_x, edge_y



import typing

import matplotlib.pyplot as plt
import pytest

import outset


@pytest.mark.parametrize(
    "impl", [outset.MarkMagnifyingGlass(), outset.mark_magnifying_glass]
)
def test_mark_magnifying_glass_single(impl: typing.Callable):
    ax = plt.gca()  # Get the current axes

    # Call the function with test values
    impl(x=0.5, y=0.5, ax=ax, color="red")

    # Save and print the path of the output image
    outpath = "/tmp/test_mark_magnifying_glass_single.png"
    plt.savefig(outpath)
    print(f"saved graphic to {outpath}")


@pytest.mark.parametrize(
    "impl", [outset.MarkMagnifyingGlass(), outset.mark_magnifying_glass]
)
def test_mark_magnifying_glass_multiple(impl: typing.Callable):
    # Multiple calls to the function with different parameters
    impl(x=0.3, y=0.3)
    impl(x=0.7, y=0.7, color="blue", color_accent="orange")

    # Save and print the path of the output image
    outpath = "/tmp/test_mark_magnifying_glass_multiple.png"
    plt.savefig(outpath)
    print(f"saved graphic to {outpath}")



from matplotlib.axes import Axes as mpl_Axes
from matplotlib.figure import Figure as mpl_Figure


def resize_figure_to_axes(fig: mpl_Figure, ax: mpl_Axes) -> None:
    """Resize a matplotlib figure to fit an axes, preserving the scale and
    position of the axes.

    Parameters
    ----------
    fig : mpl_Figure
        The matplotlib figure object to be resized.
    ax : mpl_Axes
        The matplotlib axes object the figure should be resized to fit.

    Notes
    -----
    If axis labels or titles are getting cut off, you may need to call
    `Figure.tight_layout()` before running this function.
    """
    # Save the original position of the axis in inches
    original_pos = ax.get_position()
    original_pos_in_inches = [
        original_pos.x0 * fig.get_figwidth(),
        original_pos.y0 * fig.get_figheight(),
        original_pos.width * fig.get_figwidth(),
        original_pos.height * fig.get_figheight(),
    ]

    # Get the bounding box of the axis in inches
    bbox = ax.get_tightbbox(
        renderer=fig.canvas.get_renderer(),
    ).transformed(
        fig.dpi_scale_trans.inverted(),
    )

    # Set the figure size to match the axis size
    fig.set_size_inches(
        bbox.width + 2 * bbox.x0, bbox.height + 2 * bbox.y0, forward=True
    )

    # Adjust the position of the axis back to original
    new_pos = [
        original_pos_in_inches[0] / fig.get_figwidth(),
        original_pos_in_inches[1] / fig.get_figheight(),
        original_pos_in_inches[2] / fig.get_figwidth(),
        original_pos_in_inches[3] / fig.get_figheight(),
    ]
    ax.set_position(new_pos)



import itertools as it
import typing

import frozendict
import numpy as np
from matplotlib import axes as mpl_axes
from matplotlib import colors as mpl_colors
from matplotlib import patches as mpl_patches
from matplotlib import pyplot as plt

from .compose_callout_leader_ import compose_callout_leader
from .get_vertices_extent_ import get_vertices_extent
from .make_radial_gradient_ import make_radial_gradient


def draw_callout(
    frame_xlim: typing.Tuple[float, float],
    frame_ylim: typing.Tuple[float, float],
    ax: typing.Optional[mpl_axes.Axes] = None,
    *,
    color: str = "black",
    clip_on: bool = False,
    leader_edge_kwargs: typing.Dict = frozendict.frozendict(),
    leader_face_kwargs: typing.Dict = frozendict.frozendict(),
    leader_stretch: float = 0.1,
    leader_stretch_unit: typing.Literal[
        "axes",
        "figure",
        "inches",
        "inchesfrom",
    ] = "axes",
    mark_glyph: typing.Optional[typing.Callable] = None,
    mark_glyph_kwargs: typing.Dict = frozendict.frozendict(),
    mark_retract: float = 0.1,
    zorder: float = 0,
    **kwargs,
) -> mpl_axes.Axes:
    """Annotate a rectangular region with a flyaway "zoom" indication upwards
    and to the right.

    Consists of (1) a callout leader and (2) a marked glyph. The leader is a
    gradient fill clipped to angle to a point up and to to the right of the
    framed region. The marked glyph (e.g., a numeral, an asterisk, etc.) is
    drawn at the vertex point of the leader.

    Parameters
    ----------
    frame_xlim : Tuple[float, float]
        The x-limits (xmin, xmax) of the rectangular region to be annotated.
    frame_ylim : Tuple[float, float]
        The y-limits (ymin, ymax) of the rectangular region to be annotated.
    ax : matplotlib.axes.Axes, optional
        The axes object on which to draw the callout. Defaults to `plt.gca()`.
    color : str, default "black"
        Color for the callout leader and glyph.
    clip_on : bool, default False
        Determines if drawing elements should be clipped to the axes bounding box.
    leader_edge_kwargs : Dict, default {}
        Keyword arguments for customizing the leader's edge.
    leader_face_kwargs : Dict, default {}
        Keyword arguments for customizing the leader's face.
    leader_stretch : float, default 0.1
        Size of callout leader in `leader_stretch_unit`.
    leader_stretch_unit : Literal['axes', 'figure', 'inches', 'inchesfrom'] default 'axes'
        How should leader stretch be specified?

        If 'axes' or 'figure', stretch is specified as a fraction of the axes
        or figure size, respectively. If 'inches', stretch is specified in
        inches. If 'inchesfrom', stretch is minimum necessary to place the
        marker `leader_stretch` inches from the lower left corner of the frame.
    mark_glyph : Optional[Callable], optional
        A callable to draw a glyph at the outer vertex of the callout leader.
        If None, no glyph is drawn.
    mark_retract : float, default 0.1
        Fraction to pull back glyph from the outer vertex of the callout.
    zorder : float, default 0
        Influences layer order of plot, with higher values in front.
    **kwargs
        Additional keyword arguments for matplotlib Polygon used in the callout.

    Returns
    -------
    mpl_axes.Axes
        The modified matplotlib axes object with the drawing elements added.
    """
    if ax is None:
        ax = plt.gca()

    # Draw callout leader
    ###########################################################################
    leader_vertices = compose_callout_leader(
        frame_xlim,
        frame_ylim,
        ax,
        stretch=leader_stretch,
        stretch_unit=leader_stretch_unit,
    )

    # ... outline
    underlay_patch = mpl_patches.Polygon(  # underlay
        leader_vertices,  # w/ upper right corner
        **{
            "closed": True,
            "clip_on": clip_on,
            "linewidth": 2,
            "zorder": zorder,
            **kwargs,
            **leader_edge_kwargs,
            "edgecolor": "white",
            "facecolor": "none",
            "linestyle": "-",
        },
    )
    ax.add_patch(underlay_patch)
    leader_patch = mpl_patches.Polygon(
        leader_vertices,  # w/ upper right corner
        **{
            "closed": True,
            "clip_on": clip_on,
            "linewidth": 2,
            "zorder": zorder,
            **kwargs,
            "edgecolor": color,
            "facecolor": "none",
            "linestyle": ":",
            **leader_edge_kwargs,
        },
    )
    ax.add_patch(leader_patch)

    # ... gradient fill, clipped insideleader_polygon
    img = ax.imshow(
        make_radial_gradient(),
        **{
            "alpha": 0.1,
            "aspect": "auto",
            "cmap": mpl_colors.LinearSegmentedColormap.from_list(
                "gradient",
                ["white", color],
            ),
            "extent": get_vertices_extent(leader_vertices),
            "interpolation": "nearest",
            "zorder": zorder,
            **{
                k: v
                for k, v in it.chain(kwargs.items(), leader_face_kwargs.items())
                if k not in ("linestyle",)
            },
        },
    )
    if not clip_on:
        img.set_clip_box(ax.bbox.shrunk(10, 10))  # grow axis clipping box
    img.set_clip_path(leader_patch)

    # Draw callout glyph
    ###########################################################################
    frame_upper_right = np.array([frame_xlim[1], frame_ylim[1]])
    leader_outer_vertex = tuple(np.max(np.array(leader_vertices), axis=0))
    assert len(leader_outer_vertex) == 2, leader_outer_vertex
    assert all(np.array(leader_outer_vertex) >= np.array(frame_upper_right))

    mark_coordinates = np.average(
        np.array([leader_outer_vertex, frame_upper_right]),
        weights=np.array([1.0 - mark_retract, mark_retract]),
        axis=0,
    )
    if mark_glyph is not None:
        mark_glyph(
            *mark_coordinates,
            ax=ax,
            **{
                "color": color,
                "clip_on": clip_on,
                "zorder": zorder,
                **mark_glyph_kwargs,
            },
        )

    return ax



import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns
from matplotlib.testing import decorators as mpl_testing_decorators

from outset import draw_marquee, mark_magnifying_glass, marqueeplot

# Sample data for testing
data = pd.DataFrame(
    {"x": [1, 2, 3, 4], "y": [1, 3, 2, 1], "outset": ["A", "B", "A", "B"]}
)


@mpl_testing_decorators.check_figures_equal()
def test_marqueeplot_vs_draw_marquee(fig_test: plt.Figure, fig_ref: plt.Figure):
    # Using marqueeplot
    ax_test = fig_test.subplots()
    marqueeplot(
        data, x="x", y="y", ax=ax_test, mark_glyph=mark_magnifying_glass
    )

    # Expected output using draw_marquee
    ax_ref = fig_ref.subplots()
    ax_ref.set_xlim(*ax_test.get_xlim())
    ax_ref.set_ylim(*ax_test.get_ylim())
    draw_marquee(
        (1, 4),
        (1, 3),
        ax=ax_ref,
        color=sns.color_palette()[0],
        frame_inner_pad=0.1,
        frame_outer_pad=0.0,
    )


@mpl_testing_decorators.check_figures_equal()
def test_marqueeplot_vs_draw_marquee_split(
    fig_test: plt.Figure, fig_ref: plt.Figure
):
    palette = ["green", "red"]
    # Using marqueeplot
    ax_test = fig_test.subplots()
    marqueeplot(
        data,
        x="x",
        y="y",
        mark_glyph=mark_magnifying_glass,
        outset="outset",
        outset_order=["A", "B"],
        hue="outset",
        ax=ax_test,
        frame_inner_pad=0.0,
        frame_outer_pad=0.0,
        palette=palette,
    )

    # Expected output using draw_marquee
    ax_ref = fig_ref.subplots()
    ax_ref.set_xlim(*ax_test.get_xlim())
    ax_ref.set_ylim(*ax_test.get_ylim())
    draw_marquee(
        (1, 3),
        (1, 2),
        ax=ax_ref,
        color=palette[0],
        frame_inner_pad=0.0,
        frame_outer_pad=0.0,
    )
    draw_marquee(
        (2, 4),
        (1, 3),
        ax=ax_ref,
        color=palette[1],
        frame_inner_pad=0.0,
        frame_outer_pad=0.0,
    )
    ax_ref.set_xlim(*ax_test.get_xlim())
    ax_ref.set_ylim(*ax_test.get_ylim())


def test_marqueeplot_one():
    # Create a sample dataframe
    data = pd.DataFrame({"x": [0.5], "y": [1]})
    sns.scatterplot(data=data, x="x", y="y")
    marqueeplot(
        data=data,
        x="x",
        y="y",
        frame_inner_pad=0.2,
    )
    outpath = "/tmp/test_marqueeplot_one.png"
    plt.savefig(outpath)
    print(f"saved graphic to {outpath}")


def test_marqueeplot_several():
    _fig, ax = plt.subplots(figsize=(6, 4))
    ax.set_xlim(0, 4)
    ax.set_ylim(0, 4)
    # Create sample data with a 'outset' column for grouping
    data = pd.DataFrame(
        {
            "x": [0.825, 3.1, 0.5, 1.3, 2.2, 2],
            "y": [1.2, 0.8, 3.5, 2.3, 1.1, 3.7],
            "outset": [
                "group1",
                "group1",
                "group2",
                "group2",
                "group3",
                "group3",
            ],
        }
    )

    sns.scatterplot(data=data, x="x", y="y", hue="outset", ax=ax)
    marqueeplot(
        data=data, x="x", y="y", outset="outset", ax=ax, leader_stretch=0.0
    )

    outpath = "/tmp/test_marqueeplot_several.png"
    plt.savefig(outpath)
    print(f"saved graphic to {outpath}")



import warnings

from matplotlib import axes as mpl_axes
import numpy as np

from .calc_aspect_ import calc_aspect


def set_aspect(ax: mpl_axes.Axes, aspect: float) -> None:
    """Adjust the ratio between ylim span length and xlim span length.

    The function calculates the current aspect ratio of the Axes object and
    adjusts its x-axis or y-axis limits to match the desired aspect ratio.
    If the desired aspect ratio is less than the current, the function
    increases the width of the x-axis. If it is greater, the height of the
    y-axis is increased.

    Note that axes limits are only ever widened. Axes widening is performed
    symmetrically.
    """
    (x0_, x1_), (y0_, y1_) = ax.get_xlim(), ax.get_ylim()
    before_width, before_height = np.ptp(ax.get_xlim()), np.ptp(ax.get_ylim())

    before_ratio = before_height / before_width
    ax.set_aspect(aspect, adjustable="datalim")
    after_ratio = np.ptp(ax.get_ylim()) / np.ptp(ax.get_xlim())

    # manual touch-up to ensure growth (not shrink) that is symmetrical
    if before_ratio == after_ratio:
        # The current aspect ratio matches the targeted one, no action needed
        # just need to ensure matplotlib didn't arbitrarily move axis limits
        ax.set_xlim(x0_, x1_)
        ax.set_ylim(y0_, y1_)
        assert ax.get_xlim() == (x0_, x1_) and ax.get_ylim() == (y0_, y1_)
    elif after_ratio < before_ratio:
        # plot is too tall so we need to increase the width.
        # in this case, before_height == after_height
        after_width = before_height / after_ratio
        assert after_width >= before_width
        pad = (after_width - before_width) / 2
        assert pad >= 0
        ax.set_xlim(x0_ - pad, x1_ + pad)
        ax.set_ylim(y0_, y1_)
    elif after_ratio > before_ratio:
        # plot is too wide so we need to increase the height.
        # in this case, before_width == after_width
        after_height = before_width * after_ratio
        assert after_height >= before_height
        pad = (after_height - before_height) / 2
        assert pad >= 0
        ax.set_xlim(x0_, x1_)
        ax.set_ylim(y0_ - pad, y1_ + pad)
    else:
        assert False

    # check postconditions...
    # ...targeted aspect ratio was achieved
    # ...axes limit pad-out was outwards and symmetrical
    (x0, x1), (y0, y1) = ax.get_xlim(), ax.get_ylim()
    aspect_err = aspect / calc_aspect(ax) - 1
    info = (
        f"aspect={aspect}, aspect_err={aspect_err}, "
        f"before_ratio == after_ratio {before_ratio == after_ratio}, "
        f"before_ratio={before_ratio}, after_ratio={after_ratio}, "
        f"x0_={x0_}, x1_={x1_}, y0_={y0_}, y1_={y1_}, "
        f"x0={x0}, x1={x1}, y0={y0}, y1={y1}"
    )
    assert x0 <= x0_, info
    assert y0 <= y0_, info
    assert x1 >= x1_, info
    assert y1 >= y1_, info
    assert np.isclose(x0 - x0_, x1_ - x1)
    assert np.isclose(y0 - y0_, y1_ - y1)
    if aspect_err > 0.05:
        warnings.warn(f"set_aspect {aspect_err * 100}% error")



import itertools as it
import math
import typing


def layout_corner_insets(
    num_insets: int,
    corner: typing.Literal["NE", "NW", "SE", "SW"],
    inset_grid_size: float = 0.60,
    inset_pad_ratio: float = 0.25,
) -> typing.List[typing.Tuple[float, float, float, float]]:
    """Lay out positions for `n` inset plots in a specified corner.

    The frames are filled from the chosen corner outwards, filling successive diagonals perpendicular to the cornder. Inset subfigure positions are
    ordered so that inset number increases left-to-right, top-to-bottom.

    Parameters
    ----------
    num_insets : int
        The number of inset plots to be generated.
    corner : Literal["NE", "NW", "SE", "SW"]
        The corner of the grid where the insets will be positioned.
    inset_grid_size : float, default 0.4
        The size of the grid of inset plots relative to the source plot.
    inset_pad_ratio : float, default 0.33
        Pad size between inset plots as a fraction of inset plot size.

    Returns
    -------
    List[Tuple[float, float, float, float]]
        A list of tuples, each representing `(x, y, width, height)` of an inset
        plot fractionally relative to source plot axes.

    See Also
    --------
    inset_outsets
        Manipulates plot structure to apply corner inset layout calculated by
        `layout_corner_insets`, placing outset plots in specified positions over
        the source plot axes.
    """
    dimension = int(math.ceil(math.sqrt(num_insets)))
    grid_size = inset_grid_size / dimension
    ax_size = grid_size * (1 - inset_pad_ratio)
    assert ax_size < grid_size

    # generate frame coordinates from the corner out, along diagonals
    # https://mmore500.com/2023/12/16/zigzag-traversal.html
    frame_coordinates = []
    for row, col in sorted(
        sorted(it.product(range(dimension), repeat=2), key=sum)[:num_insets],
    ):
        # Calculate the top-left corner of each frame
        x = 1.0 - grid_size - col * grid_size
        y = 1.0 - grid_size - row * grid_size
        assert 1 - inset_grid_size <= x + 1e-6
        assert x <= 1.0 - grid_size + 1e-6
        assert 1 - inset_grid_size <= y + 1e-6
        assert y <= 1.0 - grid_size + 1e-6

        if corner == "NE" or corner == "SE":
            x = grid_size * col + (grid_size - ax_size)
            assert grid_size - ax_size <= x + 1e-6
            assert x <= inset_grid_size - ax_size + 1e-6
        if corner == "SE" or corner == "SW":
            y = grid_size * row + (grid_size - ax_size)
            assert grid_size - ax_size <= y + 1e-6
            assert y <= inset_grid_size - ax_size + 1e-6

        # Add the frame to the list
        frame_coordinates.append((x, y, ax_size, ax_size))

    # take the first `num_insets` frames from corner out and then sort them
    # so that frames are ordered left-to-right, top-to-bottom
    return sorted(frame_coordinates[:num_insets], key=lambda c: (-c[1], c[0]))



import typing

import matplotlib.pyplot as plt
from matplotlib import axes as mpl_axes
from matplotlib import markers as mpl_markers

from ._auxlib.rotate_marker_ import rotate_marker


def mark_magnifying_glass(
    x: typing.Union[float, typing.Sequence[float]],
    y: typing.Union[float, typing.Sequence[float]],
    ax: typing.Optional[mpl_axes.Axes] = None,
    *,
    alpha: float = 1.0,
    color: str = "black",
    color_accent: str = "white",
    linecolor: str = "none",
    markersize: float = 15,
    rotate_angle: float = 20,
    **kwargs,
) -> None:
    """Draw magnifying glass marker(s) at specified location(s) on a matplotlib
    plot.

    This function creates a magnifying glass glyph by stacking several
    matplotlib markers.

    Parameters
    ----------
    x : float or sequence of floats
        The x-coordinate(s) where the magnifying glass(es) will be placed.
    y : float or sequence of floats
        The y-coordinate(s) where the magnifying glass(es) will be placed.
    ax : mpl_axes.Axes, optional
        The axes object on which to draw the markers.

        If None, `plt.gca()` will be used.
    color_accent : str, default "white"
        The color used for the accent layer of the glyph, enhancing visibility.
    alpha : float, default 1.0
        The transparency level of the markers.
    color : str, default "black"
        The primary color for the glyph components.
    linecolor : str, default "none"
        Color for connecting lines between markers, if any.
    markersize : float, default 15
        Size for glyph's largest marker element.
    rotate_angle : float, default 20
        The angle (in degrees) to rotate the asterisk part of the magnifying
        glass.
    **kwargs : dict, optional
        Additional keyword arguments forward to matplotlib `plot`.

    Returns
    -------
    None
    """
    if ax is None:
        ax = plt.gca()

    handle_marker = rotate_marker(3, rotate_angle)  # TICKDOWN

    # Scaling factors for different elements relative to markersize
    handle_underlay_scale = 1.5 / 1.5
    handle_scale = 1.4 / 1.5
    circle_scale = 1.3 / 1.5
    ring_scale = 1.1 / 1.5

    # Underlay handle
    ax.plot(
        x,
        y,
        alpha=alpha,
        color=linecolor,
        marker=handle_marker,
        markeredgecolor=color_accent,
        markeredgewidth=markersize / 3,
        markersize=markersize * handle_underlay_scale,
        **kwargs,
    )

    # Draw handle
    ax.plot(
        x,
        y,
        alpha=0.8 * alpha,
        color="none",  # line rendering is handled above
        marker=handle_marker,
        markeredgecolor=color,
        markeredgewidth=markersize / 6,
        markerfacecolor=color,
        markersize=markersize * handle_scale,
        **kwargs,
    )

    # Underlay glass lens
    ax.plot(
        x,
        y,
        alpha=alpha,
        color="none",  # line rendering is handled above
        marker="o",
        markeredgecolor=color_accent,
        markeredgewidth=markersize / 12,
        markerfacecolor=color_accent,
        markersize=markersize * circle_scale,
        **kwargs,
    )

    # Tint glass lens
    ax.plot(
        x,
        y,
        alpha=0.1 * alpha,
        color="none",  # line rendering is handled above
        marker="o",
        markeredgecolor="none",
        markeredgewidth=2,
        markerfacecolor=color,
        markersize=markersize * ring_scale,
        **kwargs,
    )

    # Draw glass ring
    ax.plot(
        x,
        y,
        alpha=0.7 * alpha,
        color="none",  # line rendering is handled above
        marker="o",
        markeredgecolor=color,
        markeredgewidth=markersize / 12,
        markerfacecolor="none",
        markersize=markersize * ring_scale,
        **kwargs,
    )


class MarkMagnifyingGlass:
    """Functor interface for `mark_magnifying_glass`."""

    def __call__(self, *args, **kwargs):
        """Forwards to `mark_magnifying_glass`."""
        return mark_magnifying_glass(*args, **kwargs)



import matplotlib.cbook as mpl_cbook
import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns

from outset import OutsetGrid

# Sample data for testing
data = pd.DataFrame(
    {"x": [1, 2, 3, 4], "y": [1, 3, 2, 1], "outset": ["A", "B", "A", "B"]}
)


def test_OutsetGrid_one():
    # Create a sample dataframe
    data = pd.DataFrame({"x": [0.5], "y": [1], "outset": ["A"]})
    sns.scatterplot(data=data, x="x", y="y")
    OutsetGrid(
        data=data,
        x="x",
        y="y",
        outset="outset",
        marqueeplot_kwargs={
            "frame_inner_pad": 0.2,
        },
    ).marqueeplot()
    outpath = "/tmp/test_OutsetGrid_one.png"
    plt.savefig(outpath)
    print(f"saved graphic to {outpath}")


def test_OutsetGrid_with_sourceplot_monochrome():
    # Create sample data with a 'outset' column for grouping
    data = pd.DataFrame(
        {
            "x": [0.825, 3.1, 0.5, 0.8, 2.2, 2],
            "y": [1.2, 0.8, 2.5, 2.3, 1.1, 3.7],
            "outset": [
                "group1",
                "group1",
                "group2",
                "group2",
                "group3",
                "group3",
            ],
        }
    )

    g = OutsetGrid(data=data, x="x", y="y", outset="outset")
    g.marqueeplot()
    g.map_dataframe(sns.scatterplot, x="x", y="y", hue="outset", legend=False)

    outpath = "/tmp/test_OutsetGrid_with_sourceplot_monochrome.png"
    plt.savefig(outpath)
    print(f"saved graphic to {outpath}")


def test_OutsetGrid_with_sourceplot_hue():
    # Create sample data with a 'outset' column for grouping
    data = pd.DataFrame(
        {
            "x": [0.825, 3.1, 0.5, 0.8, 2.2, 2],
            "y": [1.2, 0.8, 2.5, 2.3, 1.1, 3.7],
            "outset": [
                "group1",
                "group1",
                "group2",
                "group2",
                "group3",
                "group3",
            ],
        }
    )

    g = OutsetGrid(data=data, x="x", y="y", hue="outset")
    g.marqueeplot()
    g.map_dataframe(sns.scatterplot, x="x", y="y", legend=False)

    outpath = "/tmp/test_OutsetGrid_with_sourceplot_hue.png"
    plt.savefig(outpath)
    print(f"saved graphic to {outpath}")


def test_OutsetGrid_broadcast():
    with mpl_cbook.get_sample_data("grace_hopper.jpg") as image_file:
        image = plt.imread(image_file)

    og = OutsetGrid(
        data=[(40, 60, 60, 80), (10, 40, 14, 21)],
        col=True,
        hue=True,
    )
    og.broadcast(
        plt.imshow, image, extent=(0, 100, 0, 100), origin="upper", zorder=-1
    )
    og.marqueeplot()
    plt.savefig("/tmp/test_OutsetGrid_broadcast.png")



import itertools as it
import math
import typing

from matplotlib.transforms import Bbox as mpl_Bbox

from .OutsetGrid_ import OutsetGrid
from .layout_corner_insets_ import layout_corner_insets
from ._auxlib.resize_figure_to_axes_ import resize_figure_to_axes


def inset_outsets(
    outset_grid: OutsetGrid,
    insets: typing.Union[
        typing.Literal["NE", "NW", "SE", "SW"],
        typing.Sequence[typing.Tuple[float, float, float, float]],
    ] = "NW",
    strip_axes: bool = False,
    strip_labels: bool = True,
    strip_spines: bool = False,
    strip_ticks: bool = True,
    strip_titles: bool = True,
) -> None:
    """Rearrange OutsetGrid figure to superimpose outset axes over source axes.

    Parameters
    ----------
    outset_grid : OutsetGrid
        OutsetGrid to transform.
    insets : Union[Literal["NE", "NW", "SE", "SW"], Sequence[Tuple[float,
    float, float, float]]], deefault "NW"
        Where to place outset plots over source plot.

        If "NE", "NW", "SE", or "SW", the outset plots will be inset in the
        specified corner of the grid. Alternately, inset positioning can be
        specified directly by providing a sequence of tuples `(x0, y0, width, height)` for the position of each plot in fractional coordinates of
        sourceplot.
    strip_label : bool, default True
        Should x and y labels be stripped from inset plots.
    strip_spines : bool, default False
        Should removes the spines of the inset plots.
    strip_ticks : bool, default True
        If True, removes the ticks from the inset plots.
    strip_titles : bool, default True
        If True, removes the title from the inset plots.

    Returns
    -------
    None

    See Also
    --------
    layout_corner_insets
        Underlying engine implementing corner inset layout when "NE", "NW",
        "SE", or "SW" is passed to `inset_outsets`. The function
        `layout_corner_insets` can be called directly to provide the `insets`
        kwarg to tweak corner inset layouts.
    """
    if outset_grid.source_axes is None:
        raise ValueError("OutsetGrid missing source axes to inset outset axes")

    if isinstance(insets, str):
        insets = layout_corner_insets(outset_grid._ncol - 1, insets)

    outset_grid.figure.tight_layout()

    # Shrink axes to fit the source axes
    resize_figure_to_axes(outset_grid.figure, outset_grid.source_axes)

    # Get position and size of the source axes
    source_pos = outset_grid.source_axes.get_position()

    # Create and customize inset subplots based on the provided rectangles
    for ax, (rel_x0, rel_y0, rel_w, rel_h) in zip(
        outset_grid.outset_axes,
        insets,
        strict=True,
    ):
        # Calculate absolute position and size of each inset subplot
        abs_x0 = source_pos.x0 + rel_x0 * source_pos.width
        abs_y0 = source_pos.y0 + rel_y0 * source_pos.height
        abs_w = rel_w * source_pos.width
        abs_h = rel_h * source_pos.height

        assert -0.01 <= abs_x0 <= 1.01, abs_x0
        assert -0.01 <= abs_y0 <= 1.01, abs_y0
        assert -0.01 <= abs_w <= 1.01, abs_w
        assert -0.01 <= abs_h <= 1.01, abs_h
        assert -0.01 <= abs_x0 + abs_w <= 1.01, abs_x0 + abs_w
        assert -0.01 <= abs_y0 + abs_h <= 1.01, abs_y0 + abs_h

        # Set position of each inset subplot
        ax.set_position(mpl_Bbox.from_bounds(abs_x0, abs_y0, abs_w, abs_h))

        # Additional customization (removing labels, spines, etc.)
        if strip_axes:
            ax.set_axis_off()
        if strip_labels:
            ax.set_xlabel("")
            ax.set_ylabel("")
        if strip_spines:
            for spine in ax.spines.values():
                spine.set_visible(False)
        if strip_ticks:
            ax.tick_params(
                left=False, bottom=False, labelleft=False, labelbottom=False
            )
        if strip_titles:
            ax.set_title("")

    # Equalize aspect ratios
    outset_grid.equalize_aspect()



import typing

from matplotlib import axes as mpl_axes

from .MarkInlaidAsterisk_ import mark_inlaid_asterisk

_color_t = typing.Union[typing.Tuple, str]


class MarkAlphabeticalBadges:
    """Functor to mark sequentially-lettered badges.

    This class renders markers as letters overlaid onto a circular badge with a
    slightly larger circular underlay.
    """

    _counter: int
    _step: int

    def __init__(
        self: "MarkAlphabeticalBadges",
        start: str = "a",
        step: int = 1,
    ) -> None:
        """Initialize functor.

        Parameters
        ----------
        start : str, default "a"
            The starting number for the alphabetical badges.

            Pass "A" for uppercase letters.
        step : int, default 1
            The step size for the alphabetical badges.
        """
        self._counter = ord(start)
        self._step = step

    def __call__(
        self: "MarkAlphabeticalBadges",
        x: float,
        y: float,
        ax: typing.Optional[mpl_axes.Axes] = None,
        *,
        color: str = "black",
        color_accent: str = "white",
        color_letter: _color_t = "color_accent",
        color_badge: _color_t = "color",
        color_underlay: _color_t = "color_accent",
        letter_edgewidth: float = 1,
        marker_badge: str = "o",
        marker_underlay: str = "o",
        markersize: float = 22,
        scale_letter: float = 0.4,
        scale_badge: float = 0.8,
        **kwargs,
    ) -> None:
        """Draw lettered badge marker at specified location.

        Alphabetical values increment with each call to this functor.

        Parameters
        ----------
        x : float
            The x-coordinate where the alphabetical badge will be placed.
        y : float
            The y-coordinate where the alphabetical badge will be placed.
        ax : mpl_axes.Axes, optional
            The axes object on which to draw the marker. If None, plt.gca()
            will be used.
        color : str, default "black"
            The primary color for the glyph.
        color_letter : Union[str, tuple], default "white"
            The color of the letter.
        color_badge : Union[str, tuple], default "color"
            The color of the badge.

            If "color", primary color will be substituted.
        color_underlay : Union[str, tuple], default "color_accent"
            The color of the underlay.

            If "color_accent", accent color will be substituted.
        letter_edgewidth : float, default 1
            The edge width of the letter marker.
        marker_badge : str, default "o"
            The marker style for the badge.
        marker_underlay : str, default "o"
            The marker style for the underlay.
        markersize : float, default 22
            Size for glyph's largest marker element.
        scale_letter : float, default 0.4
            The scaling factor for the letter size.
        scale_badge : float, default 0.8
            The scaling factor for the badge size.
        **kwargs : dict, optional
            Additional keyword arguments forward to matplotlib `plot`.

        Returns
        -------
        None
        """
        mark_inlaid_asterisk(
            x=x,
            y=y,
            ax=ax,
            asterisk_edgewidth=letter_edgewidth,
            color=color,
            color_accent=color_accent,
            color_asterisk_edge=color_letter,
            color_asterisk_face=color_letter,
            color_badge=color_badge,
            color_underlay=color_underlay,
            marker=f"$\mathrm{{{chr(self._counter)}}}$",
            marker_badge=marker_badge,
            marker_underlay=marker_underlay,
            markersize=markersize,
            scale_asterisk=scale_letter,
            scale_badge=scale_badge,
            **{
                "linecolor": "none",
                **kwargs,
            },
        )
        self._counter += self._step



from matplotlib import pyplot as plt
import pandas as pd
import seaborn as sns
import outset


def test_scatter_stub():
    tips = sns.load_dataset("tips")
    outlier = pd.DataFrame(
        {
            "total_bill": [
                tips["total_bill"].mean(),
                tips["total_bill"].mean() * 8,
            ],
            "tip": [101, 200],
            "sex": [tips["sex"].mode()[0]] * 2,
            "smoker": [tips["smoker"].mode()[0]] * 2,
            "day": [tips["day"].mode()[0]] * 2,
            "time": [tips["time"].mode()[0]] * 2,
            "size": [tips["size"].mode()[0]] * 2,
        },
    )

    # Append the outlier to the original dataset
    extended_tips = pd.concat([tips, outlier], ignore_index=True)

    # Create the scatter plot
    ax = sns.scatterplot(
        data=extended_tips,
        x="total_bill",
        y="tip",
        hue="size",
        size="size",
        sizes=(20, 200),
        legend="full",
    )
    sns.despine(ax=ax)

    plt.errorbar(
        extended_tips["total_bill"],
        extended_tips["tip"],
        yerr=0.2,
        fmt="none",
        ecolor="gray",
        alpha=0.5,
    )

    outset.rescale_clip_outliers(ax, outset.CalcBoundsIQR(5))
    ax.set_xlim(0, ax.get_xlim()[1])
    ax.set_ylim(0, ax.get_ylim()[1])

    outset.stub_all_clipped_values(ax)

    plt.savefig("/tmp/test_scatter_stub.png")



import typing

from matplotlib.axes import Axes as mpl_Axes
from matplotlib.collections import PathCollection
from matplotlib import pyplot as plt
import numpy as np

from .CalcBoundsIQR_ import CalcBoundsIQR


def rescale_clip_outliers(
    ax: typing.Optional[mpl_Axes] = None,
    calc_outlier_bounds: typing.Union[
        typing.Callable,
        typing.Tuple[typing.Callable, typing.Callable],
        None,
    ] = None,
    pad: float = 0.1,
) -> None:
    """Rescale the axes of a Matplotlib axes with plotted points to exclude
    outliers.

    This function inspects plotted points already on the axes object.

    Parameters
    ----------
    ax : mpl_Axes, optional
        The Matplotlib Axes object to rescale.

        If None, plt.gca() will be used.
    calc_outlier_bounds : Union[Callable, Tuple[Callable, Callable]], optional
        A function or a tuple of two functions to calculate the outlier bounds
        for the x and y data points.

        Each function should accept an array of data points and return a tuple
        of (lower, upper) bounds. If None, defaults to interquartile
        range-based calculation (`CalcBoundsIQR` with
        multiplier 1.5).
    pad : float, default 0.1
        How far should axes limits be padded beyond the non-outlier data range?

    Returns
    -------
    None

    Notes
    -----
    In cases where there are no outliers or all points are outliers, axes limits
    will not be adjusted In cases where non-outliers take on only a single
    value, padding will be applied relative to that value or, if zero, one.
    """
    if ax is None:
        ax = plt.gca()

    if calc_outlier_bounds is None:
        calc_outlier_bounds_x = CalcBoundsIQR(iqr_multiplier=1.5)
        calc_outlier_bounds_y = CalcBoundsIQR(iqr_multiplier=1.5)
    elif isinstance(calc_outlier_bounds, tuple):
        calc_outlier_bounds_x, calc_outlier_bounds_y = calc_outlier_bounds
    elif callable(calc_outlier_bounds):
        calc_outlier_bounds_x = calc_outlier_bounds
        calc_outlier_bounds_y = calc_outlier_bounds
    else:
        raise ValueError(
            "calc_outlier_bounds must be callable or tuple, not "
            f"{calc_outlier_bounds}",
        )

    x_points, y_points = [], []

    # Gather all points from the plot
    for collection in ax.collections:
        if isinstance(collection, PathCollection):
            x_points.extend(collection.get_offsets()[:, 0])
            y_points.extend(collection.get_offsets()[:, 1])

    # Calculate IQR bounds for x and y data
    x_lower, x_upper = calc_outlier_bounds_x(x_points)
    y_lower, y_upper = calc_outlier_bounds_y(y_points)

    # Filter outliers and determine new axis limits
    x_points_filtered = [x for x in x_points if x_lower <= x <= x_upper]
    y_points_filtered = [y for y in y_points if y_lower <= y <= y_upper]

    if len(x_points_filtered):
        x0, x1 = min(x_points_filtered), max(x_points_filtered)
        xpad = pad * ((x1 - x0) or x0 or 1)
        ax.set_xlim(x0 - xpad, x1 + xpad)
        assert np.ptp(ax.get_xlim()) or pad == 0
    if len(y_points_filtered):
        y0, y1 = min(y_points_filtered), max(y_points_filtered)
        ypad = pad * ((y1 - y0) or y0 or 1)
        ax.set_ylim(y0 - ypad, y1 + ypad)
        assert np.ptp(ax.get_ylim()) or pad == 0



from matplotlib.axes import Axes as mpl_Axes


def is_axes_unset(ax: mpl_Axes) -> bool:
    """Test if axes limits have been set."""
    # see https://matplotlib.org/stable/users/faq.html#check-whether-a-figure-is-empty
    return not (
        len(ax.get_children()) > 10  # 10 objs in empty ax
        or ax.get_xlim() != (0.0, 1.0)  # in case axlim set manually...
        or ax.get_ylim() != (0.0, 1.0)
    )
